/**
 * This script implements an HTML canvas-based whiteboard.   This is based on
 * Mihai Sucan's article at http://dev.opera.com/articles/view/html5-canvas-painting/.
 * Users can draw lines, rectangles and curves.  Everything a user draws is passed
 * to an XMPP multiuser chatroom.  Similarly, drawings coming from the chatroom are
 * also redrawn on the local canvas.
 *
 * Original copyright msg:
 * Copyright 2009 ROBO Design
 * http://www.robodesign.ro
 */


/**
 * The paint object definition.
 */
function paint () {
  var canvas, context, canvasMain, contextMain, canvasRemote, contextRemote, connection;
  var x_list, y_list;

  // The active tool instance.
  var tool;
  var tool_default = 'line';

  function init () {
    // Find the canvas element.
    canvasMain = document.getElementById('imageView');
    if (!canvasMain) {
      alert('Error: I cannot find the canvas element!');
      return;
    }

    if (!canvasMain.getContext) {
      alert('Error: no canvas.getContext!');
      return;
    }

    // Get the 2D canvas context.
    contextMain = canvasMain.getContext('2d');
    if (!contextMain) {
      alert('Error: failed to getContext!');
      return;
    }


    // Create the canvas that'll capture the remote drawings.
    var container = canvasMain.parentNode;

    canvasRemote = document.createElement('canvas');
    if (!canvasRemote) {
      alert('Error: I cannot create a new remote canvas element!');
      return;
    }

    canvasRemote.id     = 'imageRemote';
    canvasRemote.width  = canvasMain.width;
    canvasRemote.height = canvasMain.height;
    container.appendChild(canvasRemote);

    contextRemote = canvasRemote.getContext('2d');


    // Add the temporary canvas.
    canvas = document.createElement('canvas');
    if (!canvas) {
      alert('Error: I cannot create a new canvas element!');
      return;
    }

    canvas.id     = 'imageTemp';
    canvas.width  = canvasMain.width;
    canvas.height = canvasMain.height;
    container.appendChild(canvas);

    context = canvas.getContext('2d');


    // Get the tool select input.
    var tool_select = document.getElementById('dtool');
    if (!tool_select) {
      alert('Error: failed to get the dtool element!');
      return;
    }
    tool_select.addEventListener('change', ev_tool_change, false);

    // Activate the default tool.
    if (tools[tool_default]) {
      tool = new tools[tool_default]();
      tool_select.value = tool_default;
    }

    // Attach the mousedown, mousemove and mouseup event listeners.
    canvas.addEventListener('mousedown', ev_canvas, false);
    canvas.addEventListener('mousemove', ev_canvas, false);
    canvas.addEventListener('mouseup',   ev_canvas, false);
  }

  // The general-purpose event handler. This function just determines the mouse 
  // position relative to the canvas element.
  function ev_canvas (ev) {
    if (ev.layerX || ev.layerX == 0) { // Firefox
      ev._x = ev.layerX;
      ev._y = ev.layerY;
    } else if (ev.offsetX || ev.offsetX == 0) { // Opera
      ev._x = ev.offsetX;
      ev._y = ev.offsetY;
    }

    // Call the event handler of the tool.
    var func = tool[ev.type];
    if (func) {
      func(ev);
    }
  }

  // The event handler for any changes made to the tool selector.
  function ev_tool_change (ev) {
    if (tools[this.value]) {
      tool = new tools[this.value]();
    }
  }


  // This object holds the implementation of each drawing tool.
  var tools = {};

  // The drawing pencil.
  tools.pencil = function () {
    var tool = this;
    this.started = false;

    // This is called when you start holding down the mouse button.
    // This starts the pencil drawing.
    this.mousedown = function (ev) {
        context.beginPath();
        context.moveTo(ev._x, ev._y);
        tool.started = true;

        tool.x0 = ev._x;
        tool.y0 = ev._y;

        // Remote stuff.
        x_list = new Array();
        y_list = new Array();

        x_list.push(ev._x);
        y_list.push(ev._y);
    };

    // This function is called every time you move the mouse. Obviously, it only 
    // draws if the tool.started state is set to true (when you are holding down 
    // the mouse button).
    this.mousemove = function (ev) {
      if (tool.started) {
        context.lineTo(ev._x, ev._y);
        context.stroke();

        // Remote stuff.
        x_list.push(ev._x);
        y_list.push(ev._y);
      }
    };

    // This is called when you release the mouse button.
    this.mouseup = function (ev) {
      if (tool.started) {
        tool.mousemove(ev);
        tool.started = false;

        img_update();
        // Remote stuff.
        var drawData = {
            type : 'pencil',
            'x_list' : x_list.toString(),
            'y_list' : y_list.toString(),
        };
        wireTransfer(drawData);
      }
    };
  };

  // The rectangle tool.
  tools.rect = function () {
    var tool = this;
    this.started = false;

    this.mousedown = function (ev) {
      tool.started = true;
      tool.x0 = ev._x;
      tool.y0 = ev._y;
    };

    this.mousemove = function (ev) {
      if (!tool.started) {
        return;
      }

      context.clearRect(0, 0, canvas.width, canvas.height);
      drawRect(tool.x0, tool.y0, ev._x, ev._y, context);
    };

    this.mouseup = function (ev) {
      if (tool.started) {
        tool.mousemove(ev);
        tool.started = false;

        img_update();

        var drawData = {
            type : 'rect',
            start_x : tool.x0,
            start_y : tool.y0,
            end_x : ev._x,
            end_y : ev._y,
        };
        wireTransfer(drawData);
      }
    };
  };

  function drawRect(x0, y0, x1, y1, contextObj)
  {
      var x = Math.min(x1, x0),
          y = Math.min(y1,  y0),
          w = Math.abs(x1 - x0),
          h = Math.abs(y1 - y0);

      if (!w || !h) {
        return;
      }

      contextObj.strokeRect(x, y, w, h);
  }


  // The line tool.
  tools.line = function () {
    var tool = this;
    this.started = false;

    this.mousedown = function (ev) {
      tool.started = true;
      tool.x0 = ev._x;
      tool.y0 = ev._y;
    };

    this.mousemove = function (ev) {
      if (!tool.started) {
        return;
      }

      context.clearRect(0, 0, canvas.width, canvas.height);
      drawLine(tool.x0, tool.y0, ev._x, ev._y, context);
    };

    this.mouseup = function (ev) {
      if (tool.started) {
        tool.mousemove(ev);
        tool.started = false;

        img_update();

        var drawData = {
            type : 'line',
            start_x : tool.x0,
            start_y : tool.y0,
            end_x : ev._x,
            end_y : ev._y,
        };
        wireTransfer(drawData);
      }
    };
  };

  function drawLine(x0, y0, x1, y1, contextObj)
  {
      contextObj.beginPath();
      contextObj.moveTo(x0, y0);
      contextObj.lineTo(x1, y1);
      contextObj.stroke();
  }


  tools.clear = function () {
    contextMain.clearRect(0, 0, canvas.width, canvas.height);

    var drawData = { type : 'clear' };
    img_update();
    wireTransfer(drawData);
  }

  function draw_curve(x_list, y_list, context_obj)
  {
      var start_x = x_list.shift();
      var start_y = y_list.shift();
      var x, y;

      context_obj.beginPath();
      context_obj.moveTo(start_x, start_y);

      for (var i in x_list)
      {
          x = x_list[i];
          y = y_list[i];

          context_obj.lineTo(x, y);
      }

      context_obj.stroke();
      context_obj.closePath();
  }


  /**
   * This function draws the #imageTemp canvas on top of #imageView, after which 
   * #imageTemp is cleared. This function is called each time when the user 
   * completes a drawing operation.
   */
  function img_update () {
    contextMain.drawImage(canvas, 0, 0);
    context.clearRect(0, 0, canvas.width, canvas.height);
  }

  function img1_update (drawData)
  {
      switch(drawData.type)
      {
          case 'line':
          drawLine(drawData.start_x, drawData.start_y, drawData.end_x, drawData.end_y, contextRemote);
          break;

          case 'rect':
          drawRect(drawData.start_x, drawData.start_y, drawData.end_x, drawData.end_y, contextRemote);
          break;

          case 'clear':
          contextMain.clearRect(0, 0, canvas.width, canvas.height);
          break;

          case 'pencil':
          draw_curve(drawData.x_list.split(','), drawData.y_list.split(','), contextRemote);
          break;

          default:
          return;
      }

      contextMain.drawImage(canvasRemote, 0, 0);
      contextRemote.clearRect(0, 0, canvasRemote.width, canvasRemote.height);
  }


  this.img2_update = function (drawData)
  {
      img1_update(drawData);
  }

  init();
}


